#include "MeshEditMode.h"
#include "Runtime/Engine/Classes/Engine/Selection.h"
#include "ProceduralMeshComponent.h"
#include "KismetProceduralMeshLibrary.h"
// (S1)

bool FMeshEditMode::IsRegistered = false;
const FEditorModeID
    FMeshEditMode::MeshEditModeID("MeshEditModeEditor");
// (S2)

// public methods
auto FMeshEditMode::Register(void)
    -> void {
    if (!IsRegistered) {
	FEditorModeRegistry::Get().RegisterMode<FMeshEditMode>
	    (MeshEditModeID); // (R1)
	IsRegistered = true;
    }
}

auto FMeshEditMode::Tick(FEditorViewportClient* ViewportClient,
			 float DeltaTime)
    -> void {
    AActor* editing_actor = nullptr;
    for (FSelectionIterator it =
	     GEditor->GetSelectedActorIterator(); it; ++it) {
	AActor* act = Cast<AActor>(*it); // (T1)
	if (act->GetComponentByClass(UProceduralMeshComponent::
				     StaticClass()) &&
	    act->GetActorLabel().
	    EndsWith("_edit", ESearchCase::CaseSensitive)) // (T2)
	{
	    editing_actor = act; // (T3)
	    break;
	} else {
	    if (GenProceduralMeshActor(act)) // (T4)
		break;
	}
    }
    if (PrevEditingActor == editing_actor) return; // (T5)

    if (editing_actor)
	AddMeshesToVertices(editing_actor); // (T6)
    if (PrevEditingActor) 
	UpdateMesh(PrevEditingActor); // (T7)

    PrevEditingActor = editing_actor; // (T8)
}

// private methods
auto FMeshEditMode::GenProceduralMeshActor(AActor* SMActor)
    -> AActor* {
    if (!SMActor) return nullptr;
    const FString pmactor_name(SMActor->GetActorLabel() +
			       "_edit"); // (G1)
    for (AActor* act : SMActor->GetLevel()->Actors)
	if (act && (act->GetActorLabel() == pmactor_name)) 
	    return nullptr; // (G2)

    UStaticMeshComponent* smc = SMActor->
	FindComponentByClass<UStaticMeshComponent>(); // (G3)
    if (!smc) return nullptr;
    UStaticMesh* smesh = smc->GetStaticMesh(); // (G4)
    smesh->bAllowCPUAccess = true; // (G5)

    TArray<FVector> vertices;
    TArray<int32> triangles;
    TArray<FVector> normals;
    TArray<FVector2D> uvs;
    TArray<FProcMeshTangent> tangents; // (G6)

    UKismetProceduralMeshLibrary::
	GetSectionFromStaticMesh(smesh, 0, 0,
				 vertices, triangles, normals,
				 uvs, tangents); // (G7)

    AActor* pm_actor =
	SMActor->GetWorld()->SpawnActor<AActor>(); // (G8)
    UProceduralMeshComponent* pmesh =
	NewObject<UProceduralMeshComponent>(pm_actor); // (G9)
    TArray<FLinearColor> vcols;
    vcols.SetNum(vertices.Num());
    for (int i = 0; i < vcols.Num(); i++)
	vcols[i] = FLinearColor::Yellow; // (G10)
    pmesh->CreateMeshSection_LinearColor(0,
			     vertices, triangles, normals,
			     uvs, vcols, tangents, true); // (G11)
    pm_actor->SetRootComponent(pmesh); // (G12)
    pmesh->RegisterComponent();
    pm_actor->SetActorLocation(SMActor->GetActorLocation());
    pm_actor->SetActorLabel(pmactor_name); // (G13)
    return pm_actor; // (G14)
}

auto FMeshEditMode::AddMeshesToVertices(AActor* PMActor)
    -> void {
    if (!PMActor) return;
    UProceduralMeshComponent* pmc = PMActor->
	FindComponentByClass<UProceduralMeshComponent>(); // (A1)
    if (!pmc) return;
    TArray<UStaticMeshComponent*> smcs;
    PMActor->GetComponents(smcs);
    for (UStaticMeshComponent* sm : smcs) {
	sm->UnregisterComponent();
	PMActor->RemoveOwnedComponent(sm); 
    } // (A2)

    UStaticMesh *mesh = LoadObject<UStaticMesh>
	(PMActor,
	 TEXT("StaticMesh'/MeshEdit/RedCube.RedCube'")); // (A3)
    const FVector scale(0.1f, 0.1f, 0.1f); // (A4)
    FProcMeshSection* msect = pmc->GetProcMeshSection(0); // (A5)
    for (int vidx = 0; vidx < msect->ProcVertexBuffer.Num();
	 vidx++) { // (A6)
	FProcMeshVertex& vert =
	    msect->ProcVertexBuffer[vidx]; // (A7)
	UStaticMeshComponent *comp =
	    NewObject<UStaticMeshComponent>(PMActor);
	comp->SetStaticMesh(mesh); // (A8)
	PMActor->AddInstanceComponent(comp);
	comp->RegisterComponent();
	comp->SetWorldScale3D(scale); // (A9)
	const FVector loc = 
	    PMActor->GetActorTransform().
	    TransformPosition(vert.Position); 
	comp->SetWorldLocation(loc); // (A10)
	comp->ComponentTags.
	    Insert(FName(*FString::FromInt(vidx)), 0); // (A11)
    }
}

auto FMeshEditMode::UpdateMesh(AActor* PMActor) 
    -> void {
    if (!PMActor) return;
    UProceduralMeshComponent* pmc = 
	PMActor->FindComponentByClass<UProceduralMeshComponent>();
    if (!pmc) return;

    FProcMeshSection* msect = pmc->GetProcMeshSection(0); // (U1)
    TArray<UStaticMeshComponent*> mesh_comps;
    PMActor->GetComponents(mesh_comps);
    for (UStaticMeshComponent* mesh_comp : mesh_comps) { // (U2)
	if (mesh_comp->ComponentTags.Num() == 0) continue; // (U3)
	const int vidx =
	    FCString::Atoi(*(mesh_comp->ComponentTags[0].
			     ToString())); // (U4)
	if (msect->ProcVertexBuffer.Num() <= vidx) return;
	FProcMeshVertex& vert =
	    msect->ProcVertexBuffer[vidx]; // (U5)
	vert.Position =
	    PMActor->GetActorTransform().InverseTransformPosition
	    (mesh_comp->GetComponentLocation()); // (U6)
    }
    pmc->SetProcMeshSection(0, *msect); // (U7)

    for (UStaticMeshComponent* mesh_comp : mesh_comps) {
	mesh_comp->UnregisterComponent();
	PMActor->RemoveOwnedComponent(mesh_comp);    
    } // (U8)
}
